# unvivtool Copyright (C) 2020-2024 Benjamin Futasz <https://github.com/bfut>
#
# You may not redistribute this program without its source code.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
  test_smoketest.py
"""
import pathlib

import pytest

# Look for local build, if not installed
try:
    import unvivtool as uvt
except ModuleNotFoundError:
    import sys
    PATH = pathlib.Path(pathlib.Path(__file__).parent / "../python/build")
    print(PATH)
    for x in PATH.glob("**"):
        sys.path.append(str(x.resolve()))
    del PATH

    import unvivtool as uvt

@pytest.mark.xfail(not hasattr(uvt, "__version__"),
                   reason="missing attribute __version__")
def test_has_attribute_version():
    if hasattr(uvt, "__version__"):
        print(f"uvt.__version__={uvt.__version__}")
    else:
        print(f'hasattr(uvt, "__version__")={hasattr(uvt, "__version__")}')
    assert hasattr(uvt, "__version__")

if __name__ == "__main__":
    print(uvt.__version__)
    test_has_attribute_version()
    help(uvt.unviv)
    print(uvt.unviv)
    print(flush=True)



# unvivtool Copyright (C) 2020-2024 Benjamin Futasz <https://github.com/bfut>
#
# You may not redistribute this program without its source code.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
  test_cli_decodeencode_c++.py
"""
import os
import pathlib
import platform
import re
import shutil
import subprocess
import struct

import pytest

def get_subprocess_ret(command, shell=False):
    ret = subprocess.run(command, shell=shell, capture_output=True)
    print("args =", ret.args)
    print("stdout =", ret.stdout.decode("utf-8"))
    print("stderr =", ret.stderr.decode("utf-8"))
#    print("Exit code", ret.returncode);
    return ret

SCRIPT_PATH = pathlib.Path(__file__).parent.resolve()
OUTDIR = SCRIPT_PATH / ".out_cli"
if platform.system() == "Windows":
    if struct.calcsize("P") * 8 == 64:
        EXECUTABLE_PATH = SCRIPT_PATH / ".." / "unvivtool_x64.exe"
    else:
        EXECUTABLE_PATH = SCRIPT_PATH / ".." / "unvivtool_x86.exe"
elif platform.system() == "Linux":
    EXECUTABLE_PATH = SCRIPT_PATH / ".." / "unvivtool_c++"
else:
    EXECUTABLE_PATH = SCRIPT_PATH / ".." / "unvivtool_macOS_c++"
VALGRIND_EXISTS = get_subprocess_ret(" ".join([ "valgrind", "--version" ]), True).returncode == 0
if VALGRIND_EXISTS:
    VALGRIND_LOG = "valgrind.log"
    EXECUTABLE_PATH = f"valgrind -s --log-file=\"{str(OUTDIR / VALGRIND_LOG)}\" " + str(EXECUTABLE_PATH)

if OUTDIR.is_dir():
    shutil.rmtree(OUTDIR)
try:
    os.mkdir(OUTDIR)
except FileExistsError:
    pass
assert OUTDIR.is_dir()

def delete_whitespace(string):
    return re.sub(r"\s+", "", string, flags=re.MULTILINE)

def get_valgrind_results(string):
    m1 = re.search(r"^.*?in use at exit:.*?blocks\n", string, flags=re.MULTILINE)
    m2 = re.search(r"^.*?total heap usage:.*?allocated\n", string, flags=re.MULTILINE)
    m3 = re.search(r"^.*?ERROR SUMMARY:.*?\n", string, flags=re.MULTILINE)
    return m1.group(0)[:-1], m2.group(0)[:-1], m3.group(0)[:-1]

def delete_file(path):
    try:
        os.remove(path)
    except FileNotFoundError:
        pass

# ------------------------------------------------------------------------------
def test_dir():
    print(OUTDIR)
    print(EXECUTABLE_PATH)
    if platform.system() == "Windows":
        ret = get_subprocess_ret(" ".join([
            "dir",
        ]), True)
    else:
        ret = get_subprocess_ret(" ".join([
            "ls", "-lg",
        ]), True)
    assert OUTDIR.exists()
    assert True

@pytest.mark.skipif(0, reason="")
def test_cli_smoketest():
    err = ""
    out = \
"""Usage: unvivtool d [<options>...] <path/to/input.viv> [<path/to/existing/output_directory>]
       unvivtool e [<options>...] <path/to/output.viv> <paths/to/input_files>...
       unvivtool <path/to/input.viv>
       unvivtool <paths/to/input_files>...

Commands:
  d             Decode and extract files from VIV/BIG archive
  e             Encode files in new VIV/BIG archive

Options:
  -aot          decoder Overwrite mode: auto rename existing file
  -dnl<N>       decode/encode, set fixed Directory eNtry Length (<N> >= 10)
  -i<N>         decode file at 1-based Index <N>
  -f<name>      decode File <name> (cAse-sEnsitivE) from archive, overrides -i
  -fh           decode/encode to/from Filenames in Hexadecimal
  -fmt<format>  encode to Format 'BIGF' (default), 'BIGH' or 'BIG4'
  -p            Print archive contents, do not write to disk (dry run)
  -we           Write re-Encode command to path/to/input.viv.txt (keep files in order)
  -v            print archive contents, Verbose
"""

    ret = get_subprocess_ret(" ".join([f"{EXECUTABLE_PATH}"]), True)
    stdout = delete_whitespace(ret.stdout.decode("utf-8"))
    assert ret.returncode == 0
    if "valgrind" not in str(EXECUTABLE_PATH):
        assert ret.stderr.decode("utf-8") == err
    else:
        log = (OUTDIR / VALGRIND_LOG).read_text(encoding="utf-8")
        a, b, c = get_valgrind_results(log)
        print(a)
        print(b)
        print(c)
        # print(log)
    assert re.sub(r"^.*?Usage", "Usage", stdout) == delete_whitespace(out)



@pytest.mark.skipif(0, reason="")
@pytest.mark.parametrize("verbose, dry",
    [ ("", ""),
      ("-v", ""),
      ("", "-p"),
      ("-v", "-p"), ])
def test_cli_decode_all_existing_dir(verbose, dry):
    err = ""
    out = \
"""Directory Entries (header) = 2 (0x2)
Header Size (header) = 55 (0x37)
File format (parsed) = BIGF
Directory Entries (parsed) = 2
Number extracted: 2
Decoder successful.
"""
    outv = \
"""Printing archive directory:

   id Valid       Offset          Gap         Size Len  Name
 ---- ----- ------------ ------------ ------------ ---  -----------------------
                       0                        55      header
 ---- ----- ------------ ------------ ------------ ---  -----------------------
    1     1           55            0        35149   8  LICENSE
    2     1        35204            0          103  15  pyproject.toml
 ---- ----- ------------ ------------ ------------ ---  -----------------------
                   35307                     35252      2 files
Number extracted: 2
Decoder successful.
"""
    poutv = \
"""Printing archive directory:

   id Valid       Offset          Gap         Size Len  Name
 ---- ----- ------------ ------------ ------------ ---  -----------------------
                       0                        55      header
 ---- ----- ------------ ------------ ------------ ---  -----------------------
    1     1           55            0        35149   8  LICENSE
    2     1        35204            0          103  15  pyproject.toml
 ---- ----- ------------ ------------ ------------ ---  -----------------------
                   35307                     35252      2 files
End dry run
Decoder successful.
"""
    delete_file(OUTDIR / "LICENSE")
    delete_file(OUTDIR / "pyproject.toml")
    ret = get_subprocess_ret(" ".join([
      f"{EXECUTABLE_PATH}",
      "d",
      verbose, dry,
      f"{SCRIPT_PATH / 'in/car.viv'}",
      f"{OUTDIR}",  # _{verbose}
    ]), True)
    stdout = delete_whitespace(ret.stdout.decode("utf-8"))
    assert ret.returncode == 0
    if "valgrind" not in str(EXECUTABLE_PATH):
        assert ret.stderr.decode("utf-8") == err
    else:
        log = (OUTDIR / VALGRIND_LOG).read_text(encoding="utf-8")
        a, b, c = get_valgrind_results(log)
        print(a)
        print(b)
        print(c)
        # print(log)
    if dry != "-p":
        if verbose != "-v":
            assert re.sub(r"^.*?Directory", "Directory", stdout) == delete_whitespace(out)
        else:
            assert re.sub(r"^.*?Printing", "Printing", stdout) == delete_whitespace(outv)
    else:
        assert re.sub(r"^.*?Printing", "Printing", stdout) == delete_whitespace(poutv)

# @pytest.mark.skipif(0, reason="")
# @pytest.mark.parametrize("verbose, dry",
#     [ ("", "", ""),
#       ("-v", "", ""),
#       ("", "", "-p"),
#       ("-v", "", "-p"), ])
# def test_cli_decode_existing_file_existing_dir(verbose, dry):
#     err = ""
#     out = \
# """Archive Size (parsed) = 35307 (0x89eb)
# Directory Entries (header) = 2
# Decoder successful.
# """
#     outv = \
# """Archive Size (parsed) = 35307 (0x89eb)
# Archive Size (header) = 35307 (0x89eb)
# Header Size (header) = 55 (0x37)
# File format (parsed) = BIGF
# Directory Entries (header) = 2
# Directory Entries (parsed) = 2
# Requested file idx = 2
# Requested file = pyproject.toml
# Buffer = 4096
# Header Size (parsed) = 55 (0x37)

# Printing VIV directory:

#    id       Offset Gap         Size Len  Name
#  ---- ------------ --- ------------ ---  -----------------------
#     1           55   0        35149   8  LICENSE
#     2        35204   0          103  15  pyproject.toml
#  ---- ------------ --- ------------ ---  -----------------------
#              35307            35252      2 files
# Decoder successful.
# """
#     poutv = \
# """Archive Size (parsed) = 35307
# Directory Entries (header) = 2
# Archive Size (header) = 35307
# Header Size (header) = 55 (0x37)
# File format (parsed) = BIGF
# Directory Entries (parsed) = 2
# Requested file idx = 2
# Requested file = pyproject.toml
# Buffer = 4096
# Header Size (parsed) = 55

# Printing VIV directory:

#    id       Offset Gap         Size Len  Name
#  ---- ------------ --- ------------ ---  -----------------------
#     1           55   0        35149   8  LICENSE
#     2        35204   0          103  15  pyproject.toml
#  ---- ------------ --- ------------ ---  -----------------------
#              35307            35252      2 files
# End dry run
# Decoder successful.
# """
#     ret = get_subprocess_ret(" ".join([
#       f"{EXECUTABLE_PATH}",
#       "d",
#       verbose, dry,
#       "-fn pyproject.toml",
#       f"{SCRIPT_PATH / 'in/car.viv'}",
#       f"{OUTDIR}",
#     ]), True)
#     assert ret.returncode == 0
#     assert ret.stderr.decode("utf-8") == err
#     if dry != "-p":
#         if verbose != "-v":
#             assert re.sub(r"^.*?\n+.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#         else:
#             assert re.sub(r"^.*?\n+.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     else:
#         assert ret.stderr.decode("utf-8") == "" #err
#         assert re.sub(r"^.*?\n+.*?\n.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(poutv)


# @pytest.mark.skipif(0, reason="")
# @pytest.mark.parametrize("verbose, dry",
#     [ ("", "", ""),
#       ("-v", "", ""),
#       ("", "", "-p"),
#       ("-v", "", "-p"), ])
# def test_cli_decode_not_existing_file_existing_dir(verbose, dry):
#     err = \
# """Cannot find requested file in archive (filename is cAse-sEnsitivE)
# """
#     out = \
# """Archive Size (parsed) = 35307
# Directory Entries (header) = 2
# Decoder failed.
# """
#     outv = out  # ends on error, hence no more infp
#     poutv = outv  # ends on error, hence no "Endr dry run" message
#     ret = get_subprocess_ret(" ".join([
#       f"{EXECUTABLE_PATH}",
#       "d",
#       verbose, dry,
#       "-fn not_in_archive",
#       f"{SCRIPT_PATH / 'in/car.viv'}",
#       f"{OUTDIR}",
#     ]), True)
#     assert ret.returncode != 0
#     assert ret.stderr.decode("utf-8") == err
#     if dry != "-p":
#         if verbose != "-v":
#             assert re.sub(r"^.*?\n+.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#         else:
#             assert re.sub(r"^.*?\n+.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     else:
#         assert re.sub(r"^.*?\n+.*?\n+.*?\n+.*?\n+.*?\n+Archive", "Archive", ret.stdout.decode("utf-8")) == delete_whitespace(poutv)


# @pytest.mark.skipif(0, reason="")
# @pytest.mark.parametrize("verbose, strict, dry",
#     [ ("", "", ""),
#       ("-v", "", ""),
#       ("", "-strict", ""),
#       ("-v", "-strict", ""),# ])
#       ("", "", "-p"),
#       ("-v", "", "-p"),
#       ("", "-strict", "-p"),
#       ("-v", "-strict", "-p"), ])
# def test_cli_encode_with_existing(verbose, strict, dry):
#     err = ""
#     out = \
# """Number of files to encode = 2
# Encoder successful.
# """
#     outv = \
# """Number of files to encode = 2
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# Encoder successful.
# """
#     poutv = \
# """Number of files to encode = 2
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# End dry run
# Encoder successful.
# """
#     ret = get_subprocess_ret(" ".join([
#       f"{EXECUTABLE_PATH}",
#       "e",
#       verbose, strict, dry,
#       f"{OUTDIR / 'car.viv'}",
#       f"{SCRIPT_PATH / 'in/LICENSE'}", f"{SCRIPT_PATH / 'in/pyproject.toml'}",
#     ]), True)
#     assert ret.returncode == 0
#     assert ret.stderr.decode("utf-8") == err
#     assert ret.returncode == 0
#     if dry != "-p":
#       assert ret.stderr.decode("utf-8") == err
#       if verbose != "-v":
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#       else:
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     else:
#       assert ret.stderr.decode("utf-8") == "" # no error
#       assert re.sub(r"^.*?\n+.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(poutv)
#     """
#     if verbose != "-v":
#         assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#     else:
#         assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     """


# @pytest.mark.skipif(0, reason="")
# @pytest.mark.parametrize("verbose, strict, dry",
#     [ ("", "", ""),
#       ("-v", "", ""),
#       ("", "-strict", ""),
#       ("-v", "-strict", ""),# ])
#       ("", "", "-p"),
#       ("-v", "", "-p"),
#       ("", "-strict", "-p"),
#       ("-v", "-strict", "-p"), ])
# def test_cli_encode_with_skip(verbose, strict, dry):
#     err = ""
#     out = \
# """Number of files to encode = 3
# Cannot open file. Skipping 'not_a_file'
# Encoder successful.
# """
#     outv = \
# """Number of files to encode = 3
# Cannot open file. Skipping 'not_a_file'
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# Encoder successful.
# """
#     poutv = \
# """Number of files to encode = 3
# Cannot open file. Skipping 'not_a_file'
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# End dry run
# Encoder successful.
# """
#     ret = get_subprocess_ret(" ".join([
#       f"{EXECUTABLE_PATH}",
#       "e",
#       verbose, strict, dry,
#       f"{OUTDIR / 'car.viv'}",
#       f"{SCRIPT_PATH / 'in/LICENSE'}", "not_a_file", f"{SCRIPT_PATH / 'in/pyproject.toml'}",
#     ]), True)
#     assert ret.returncode == 0
#     assert ret.stderr.decode("utf-8") == err
#     """
#         if verbose != "-v":
#             assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#         else:
#             assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     """
#     assert ret.returncode == 0
#     if dry != "-p":
#       assert ret.stderr.decode("utf-8") == err
#       if verbose != "-v":
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#       else:
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     else:
#       assert ret.stderr.decode("utf-8") == "" #err
#       assert re.sub(r"^.*?\n+.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(poutv)


# @pytest.mark.skipif(0, reason="")
# @pytest.mark.parametrize("verbose, strict, dry",
#     [ ("", "", ""),
#       ("-v", "", ""),
#       ("", "-strict", ""),
#       ("-v", "-strict", ""),# ])
#       ("", "", "-p"),
#       ("-v", "", "-p"),
#       ("", "-strict", "-p"),
#       ("-v", "-strict", "-p"), ])
# def test_cli_encode_to_invalid_with_existing(verbose, strict, dry):
#     err = \
# """viv: Cannot create output file 'not_a_dir/car.viv'
# """
#     out = \
# """Number of files to encode = 2
# Encoder failed.
# """
#     outv = \
# """Number of files to encode = 2
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# Encoder failed.
# """
#     poutv = \
# """Number of files to encode = 2
# Buffer = 4096
# Header Size = 55
# Directory Entries = 2
# Archive Size = 35307

#    id       Offset         Size Len  Name
#  ---- ------------ ------------ ---  -----------------------
#     1           55        35149   8  LICENSE
#     2        35204          103  15  pyproject.toml
#  ---- ------------ ------------ ---  -----------------------
#              35307        35252      2 files
# End dry run
# Encoder successful.
# """
#     ret = get_subprocess_ret(" ".join([
#       f"{EXECUTABLE_PATH}",
#       "e",
#       verbose, strict, dry,
#       f"{'not_a_dir/' + 'car.viv'}",
#       f"{SCRIPT_PATH / 'in/LICENSE'}", f"{SCRIPT_PATH / 'in/pyproject.toml'}",
#     ]), True)
#     if dry != "-p":
#       assert ret.returncode != 0
#       assert ret.stderr.decode("utf-8") == err
#       if verbose != "-v":
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(out)
#       else:
#           assert re.sub(r"^.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(outv)
#     else:
#       assert ret.returncode == 0
#       assert ret.stderr.decode("utf-8") == "" #err
#       assert re.sub(r"^.*?\n+.*?\n+.*?\n+Number", "Number", ret.stdout.decode("utf-8")) == delete_whitespace(poutv)




# unvivtool Copyright (C) 2020-2024 Benjamin Futasz <https://github.com/bfut>
#
# You may not redistribute this program without its source code.
# README.md may not be removed or altered from any unvivtool redistribution.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
setup.py - adapted from https://github.com/pybind/python_example/blob/master/setup.py
"""
import os
import platform
import pathlib
import setuptools

if "UVTVERBOSE" in os.environ:
    print(f'UVTVERBOSE={os.environ["UVTVERBOSE"]}')

script_path = pathlib.Path(__file__).parent.resolve()
os.chdir(script_path)

print("try reading VERSION_INFO from libnfsviv.h...")
with open(script_path / "./libnfsviv.h", mode="r", encoding="utf8") as f:
    for _ in range(80 - 1):
        next(f)
    __version__ = f.readline()
    print(f"readline() yields '{__version__}'")
    __version__ = __version__.rstrip().split("\"")[-2]
    print(f"VERSION_INFO={__version__}")
long_description = (script_path / "./python/README.md").read_text(encoding="utf-8")

extra_compile_args = []
if "PYMEM_MALLOC" in os.environ:
    print(f'PYMEM_MALLOC={os.environ["PYMEM_MALLOC"]}')
    extra_compile_args += [ "-DPYMEM_MALLOC" ]
if platform.system() == "Windows":
    extra_compile_args += [
        ("/wd4267"),  # prevents warnings on conversion from size_t to int
        ("/std:c++latest"), ("/Zc:__cplusplus"),  # sets __cplusplus
    ]
else:
    extra_compile_args += [
        # debug
        # ("-g"),
        # ("-pedantic-errors"),  # multi-phase extension gives error
        ("-fvisibility=hidden"),  # sets the default symbol visibility to hidden
        ("-Wformat-security"),
        ("-Wdeprecated-declarations"),
    ]

    if "gcc" in platform.python_compiler().lower():
        extra_compile_args += [
            # ("-Wno-sign-compare"),  # prevents warnings on conversion from size_t to int
            ("-Wno-unused-parameter"),  # self in PyObject *unviv(PyObject *self, ...)
            ("-Wno-missing-field-initializers"),  # {NULL, NULL} in struct PyMethodDef{}

            ("-fasynchronous-unwind-tables"),
            ("-Wall"),
            ("-Wextra"),
            ("-Wstack-protector"),
            ("-Woverlength-strings"),
            ("-Wpointer-arith"),
            ("-Wunused-local-typedefs"),
            ("-Wunused-result"),
            ("-Wvarargs"),
            ("-Wvla"),
            ("-Wwrite-strings"),  # ("-Wno-discarded-qualifiers"),

            # # https://developers.redhat.com/blog/2018/03/21/compiler-and-linker-flags-gcc
            ("-D_GNU_SOURCE"),
            ("-D_FORTIFY_SOURCE=2"),
            # ("-D_GLIBCXX_ASSERTIONS"),
            # ("-fexceptions"),  # C++ exceptions
            # ("-fplugin=annobin"),
            ("-fstack-clash-protection"),
            ("-fstack-protector-strong"),
            # ("-mcet"), ("-fcf-protection"),  # (x86 only)
            ("-Werror=format-security"),
            ("-Werror=implicit-function-declaration"),
            # ("-Wl,-z,defs"),
            # ("-Wl,-z,now"),
            # ("-Wl,-z,relro"),
        ]
    elif "clang" in platform.python_compiler().lower():
        extra_compile_args += [
            # ("-Weverything"),
            ("-Wno-braced-scalar-init"),
            # ("-Wno-newline-eof"),

            ("-D_GNU_SOURCE"),
        ]

ext_modules = [
    setuptools.Extension(
        name="unvivtool",
        sources=sorted(["./python/unvivtoolmodule.c"]),
        define_macros=[
            ("VERSION_INFO", __version__),
            ("UVTVERBOSE", os.environ.get("UVTVERBOSE", 0)),  # 0 if key not set
            ("UVTUTF8", os.environ.get("UVTUTF8", 0)),  # branch: unviv() detects utf8
        ],
        extra_compile_args=extra_compile_args,
    )
]

setuptools.setup(
    name="unvivtool",
    version=__version__,
    author="Benjamin Futasz",
    url="https://github.com/bfut/unvivtool",
    license="GPLv3+",
    description="VIV/BIG decoder/encoder",
    long_description=long_description,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    python_requires=">=3.9",
    # extras_require={"test": "pytest"},
    # Currently, build_ext only provides an optional "highest supported C++
    # level" feature, but in the future it may provide more features.
    # cmdclass={"build_ext": build_ext},
    zip_safe=False,
)



# fcecodec Copyright (C) 2020-2024 Benjamin Futasz <https://github.com/bfut>
#
# You may not redistribute this program without its source code.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
  test_decodeencode.py
"""
import os
import pathlib
import platform
import shutil
import sys

import pytest

# --------------------------------------
script_path = pathlib.Path(__file__).parent
invivfile = script_path / "in/car.viv"
outdir = script_path / ".out"
request_fileid = 2            # optional
request_filename = "LICENSE"  # optional, overrides request_fileid

if outdir.is_dir():
    shutil.rmtree(outdir)
elif outdir.exists():
    raise FileExistsError(outdir)
try:
    os.mkdir(outdir)
except FileExistsError:
    pass

if platform.python_implementation() != "PyPy" and __name__ != "__main__":
    import tracemalloc
    # tracemalloc -- BEGIN ------------------------------------
    # tracemalloc.stop()

    first_size, first_peak = tracemalloc.get_traced_memory()

    if sys.version_info[0:2] >= (3, 9):
        tracemalloc.reset_peak()

    tracemalloc.start(2)
    # tracemalloc -- END --------------------------------------

# Look for local build, if not installed
try:
    import unvivtool as uvt
except ModuleNotFoundError:
    import sys
    PATH = pathlib.Path(pathlib.Path(__file__).parent / "../python/build")
    print(PATH)
    for x in PATH.glob("**"):
        sys.path.append(str(x.resolve()))
    del PATH

    import unvivtool as uvt

if platform.python_implementation() != "PyPy" and __name__ != "__main__":
    # tracemalloc -- BEGIN ------------------------------------
    # Source: https://docs.python.org/3/library/tracemalloc.html
    import linecache
    import os
    import tracemalloc

    # https://docs.python.org/3/library/tracemalloc.html#filter
    def display_top(snapshot, key_type="lineno", limit=10):
        snapshot = snapshot.filter_traces((
            tracemalloc.Filter(True, os.__file__, all_frames=True),
            tracemalloc.Filter(True, pathlib.__file__, all_frames=True),
            tracemalloc.Filter(False, platform.__file__, all_frames=True),
            tracemalloc.Filter(False, shutil.__file__, all_frames=True),
            tracemalloc.Filter(False, tracemalloc.__file__, all_frames=True),
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            # tracemalloc.Filter(True, "<unknown>", all_frames=True),

            tracemalloc.Filter(True, pytest.__file__, all_frames=True),
            tracemalloc.Filter(True, uvt.__file__, all_frames=True),
        ))
        top_stats = snapshot.statistics(key_type)

        # assert len(top_stats) < limit
        print(tracemalloc.get_traceback_limit())


        print(f"Top {limit} lines")
        for index, stat in enumerate(top_stats[:limit], 1):
            for i in range(len(stat.traceback)):
                frame = stat.traceback[0]
                print(f"#{index}: {frame.filename}:{frame.lineno}: "
                      f"{(stat.size / 1024):.1f} KiB "
                      f"({stat.size}) "
                      f"{i}")
                line = linecache.getline(frame.filename, frame.lineno).strip()
                if line:
                    print(f"    {line}")

        other = top_stats[limit:]
        if other:
            size = sum(stat.size for stat in other)
            print(f"{len(other)} other: {(size / 1024):.1f} KiB" ,
                f"({size})")

        total = sum(stat.size for stat in top_stats)
        print(f"Total allocated size: {(total / 1024):.1f} KiB ({total})")
    # tracemalloc -- END --------------------------------------

def test_decode1():
    print("Expected result: extract all, return 1")
    retv = uvt.unviv(invivfile, outdir, verbose=1)
    assert retv == 1

def test_decode2():
    print("Expected result: extract LICENSE, return 1")
    retv = uvt.unviv(invivfile, outdir, fileidx=0, filename=request_filename, verbose=1, dry=False)
    assert retv == 1

def test_decode3():
    print("Expected result: extract file at index 2, return 1")
    retv = uvt.unviv(invivfile, outdir, fileidx=request_fileid, verbose=True)
    assert retv == 1

def test_decode4():
    print("Expected result: extract LICENSE, return 1")
    retv = uvt.unviv(invivfile, outdir, fileidx=request_fileid, filename=request_filename, verbose=1)
    assert retv == 1

def test_decode5():
    print("Expected result: outdir does not exist and is created, return 1")
    try:
        os.rmdir(outdir / "not_a_dir")
    except FileNotFoundError:
        pass
    retv = uvt.unviv(invivfile, outdir / "not_a_dir", verbose=1)
    assert retv == 1

def test_decode6():
    print("Expected result: extract LICENSE, return 1")
    retv = uvt.unviv(invivfile, outdir, filename=request_filename, verbose=1)
    assert retv == 1

def test_decode7():
    print("Expected result: extract file at index 2, return 1")
    retv = uvt.unviv(invivfile, outdir, fileidx=request_fileid, verbose=1)
    assert retv == 1

@pytest.mark.xfail(sys.platform.startswith("win"),
                   reason="encoding issues on Windows...")
def test_decode8():
    retv = uvt.unviv(script_path / "in/@二.viv", outdir, verbose=1, filename=request_filename)
    assert retv == 1

def test_decode9():
    print('Test9: uvt.unviv(invivfile, outdir, keyword="ß二")'.encode())
    print("Expected result: cannot find requested file, return 0")
    retv = uvt.unviv(invivfile, outdir, filename="ß二", verbose=1)
    assert retv == 0

def test_decode10():
    print('Test10: uvt.unviv("öäü", None)')
    print("Expected result: TypeError")
    try:
        uvt.unviv("öäü", None)
        assert False
    except TypeError as e:
        print("TypeError:", e)

def test_decode11():
    print('Test11: uvt.unviv("foo", None)')
    print("Expected result: TypeError")
    try:
        uvt.unviv("foo", None)
        assert False
    except TypeError as e:
        print("TypeError:", e)


def test_encode1():
    vivfile = script_path / ".out/car_out_encode1.viv"
    infiles = [ str(script_path / "in/LICENSE"), str(script_path / "in/pyproject.toml") ]  # (7+1) + (14+1) = 23
    for path in infiles:
        assert pathlib.Path(path).is_file()
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode2():
    print('Test2: infiles = [ "in/LICENSE", "invalid_path", "in/pyproject.toml" ]')
    print("Expected result: skip path that does not exist, encode the rest, return 1", flush=True)
    vivfile = script_path / ".out/car_out_encode2.viv"
    infiles = [ "in/LICENSE", "invalid_path", "in/pyproject.toml" ]  # (7+1) + (12+1) + (14+1) = 36
    infiles = [str(script_path / path) for path in infiles]
    for path in infiles:
        assert pathlib.Path(path).exists() or pathlib.Path(path).stem == "invalid_path"
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode3():
    print("Test3: infiles = []")
    vivfile = script_path / ".out/car_out_encode3.viv"
    infiles = []
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and len(vivfile.read_bytes()) == 0

def test_encode4():
    print("Test4: infiles = [infiles, infiles]")
    print("Expected result: TypeError", flush=True)
    vivfile = script_path / ".out/car_out_encode4.viv"
    infiles = []
    infiles = [infiles, infiles]
    assert isinstance(infiles, list)
    try:
        uvt.viv(vivfile, infiles)
        assert False
    except TypeError as e:
        print("TypeError:", e)
        assert str(e) == "expected list of str"
    assert not vivfile.exists()

@pytest.mark.xfail(sys.platform.startswith("win"),
                   reason="encoding issues on Windows...")
def test_encode5():
    print('Test5: vivfile = "tests/@二.viv"'.encode())
    print("Expected result: Linux - encode all, return 1")
    vivfile = script_path / ".out/@二_encode5.viv"
    infiles = [ str(script_path / "in/LICENSE"), str(script_path / "in/pyproject.toml") ]
    for path in infiles:
        assert pathlib.Path(path).is_file()
    retv = uvt.viv(vivfile, infiles, verbose=True)
    assert retv == 1 and vivfile.exists()

def test_encode6():
    print('Test6: infiles = ["in/LICENSE", "in/pyproject.toml", "in"]')
    print("Expected result: encode existing files, skip existing directory, return 1")
    vivfile = script_path / ".out/car_out_encode6.viv"
    infiles = [ str(script_path / "in/LICENSE"), str(script_path / "in/pyproject.toml"), str(script_path / "in/") ]
    for path in infiles:
        assert pathlib.Path(path).is_file() or pathlib.Path(path).is_dir()
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode7():
    print('Test7: infiles = ["in/LICENSE", "in/pyproject.toml", "."]')
    print("Expected result: encode existing files, skip existing directory, return 1")
    vivfile = script_path / ".out/car_out_encode7.viv"
    infiles = [ str(script_path / "in/LICENSE"), str(script_path / "in/pyproject.toml"), str(script_path / "./") ]
    for path in infiles:
        assert pathlib.Path(path).is_file() or pathlib.Path(path).is_dir()
    retv = uvt.viv(vivfile, infiles, verbose=True)
    assert retv == 1 and vivfile.exists()

def test_encode8():
    print('Test8: infiles = ["in/LICENSE", "in/pyproject.toml", "in/ß二", "in/öäü"]'.encode())
    print("Expected result: encode existing files, return 1")
    vivfile = script_path / ".out/car_out_encode8.viv"
    infiles = [ "in/LICENSE", "in/pyproject.toml", "in/ß二" ]
    infiles = [str(script_path / path) for path in infiles]
    for path in infiles:
        assert pathlib.Path(path).is_file()
    retv = uvt.viv(vivfile, infiles, verbose=True)
    assert retv == 1 and vivfile.exists()

def test_encode9():
    vivfile = script_path / ".out/car_out_encode9.viv"
    infiles = ( "in/LICENSE", "in/pyproject.toml" )
    assert isinstance(infiles, tuple)
    try:
        uvt.viv(vivfile, infiles)
        assert False
    except TypeError as e:
        print("TypeError:", e)
        assert str(e) == "expected list"
    assert not vivfile.exists()

def test_encode10():
    vivfile = script_path / ".out/car_out_encode10.viv"
    infiles = None
    try:
        uvt.viv(vivfile, infiles)
        assert False
    except TypeError as e:
        print("TypeError:", e)
        assert str(e) == "expected list"
    assert not vivfile.exists()

def test_encode11():
    vivfile = None
    infiles = [ "in/LICENSE", "in/pyproject.toml" ]
    try:
        uvt.viv(vivfile, infiles)
        assert False
    except TypeError as e:
        print("TypeError:", e)
        assert str(e) == "expected str, bytes or os.PathLike object, not NoneType"

def test_encode12():
    vivfile = script_path / ".out/car_out_encode12.viv"
    infiles = [ "invalid_path", "invalid_path", "invalid_path" ]
    infiles = [str(script_path / path) for path in infiles]
    print(infiles)
    for path in infiles:
        assert not pathlib.Path(path).exists()
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode13():
    vivfile = script_path / ".out/car_out_encode13.viv"
    infiles = [ "foo/invalid_path1", "bar/invalid_path2", "in/invalid_path3" ]  # notice no file-ending
    print(infiles)
    for path in infiles:
        assert not pathlib.Path(path).exists()
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode14():
    vivfile = script_path / ".out/car_out_encode14.viv"
    infiles = [ "invalid_path", "invalid_path", "invalid_path" ]
    print(script_path, infiles)
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode15():
    vivfile = script_path / ".out/car_out_encode15.viv"
    infiles = [ "in/LICENSE", "in/pyproject.toml", "invalid_path" ]  # notice file-ending
    print(script_path, infiles)
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists()

def test_encode16():
    vivfile = script_path / ".out/car_out_encode16.viv"
    infiles = [ "in/LICENSE", "in/pyproject.toml", ]
    print(script_path, infiles)
    retv = uvt.viv(vivfile, infiles, verbose=1, dry=0)
    assert retv == 1 and vivfile.exists()

def test_encode17():
    """uvt.viv() overwrite existing file"""
    vivfile = script_path / ".out/car_out_encode17.viv"
    infiles = [ "invalid" ]
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists() and len(vivfile.read_bytes()) == 16
    infiles = [ str(script_path / "in/LICENSE"), str(script_path / "in/pyproject.toml") ]
    for path in infiles:
        assert pathlib.Path(path).is_file()
    retv = uvt.viv(vivfile, infiles, verbose=1)
    assert retv == 1 and vivfile.exists() and len(vivfile.read_bytes()) > 16


@pytest.mark.skipif(platform.python_implementation() == "PyPy",
                   reason="'pypy-3.8' no tracemalloc")
@pytest.mark.xfail(sys.platform != "linux",
                   reason="encoding issues on Windows...")
def test_tracemalloc1():
    # tracemalloc -- BEGIN ------------------------------------
    # tracemalloc.stop()
    second_size, second_peak = tracemalloc.get_traced_memory()
    # tracemalloc.start()

    snapshot = tracemalloc.take_snapshot()
    display_top(snapshot, limit=40)

    print(f"first_size={first_size}", f"first_peak={first_peak}")
    print(f"second_size={second_size}", f"second_peak={second_peak}")
    print(flush = True)
    # tracemalloc -- END --------------------------------------

    # macos   second_size=507913  second_peak=1843124
    # ubuntu  second_size=245752  second_peak=284290
    # windows second_size=578873  second_peak=1866884
    assert second_size < 580000 and second_peak < 1870000

if __name__ == "__main__":
    print(uvt.__version__)
    test_encode1()
    test_encode2()
    test_encode3()
    test_encode4()
    test_encode5()
    test_encode6()
    test_encode7()
    test_encode8()
    test_encode9()
    test_encode10()
    test_encode11()
    test_encode12()
    test_encode13()
    test_encode14()
    test_encode15()
    test_encode16()
    test_encode17()
    pass
